import subprocess as subp
import shutil
from pathlib import Path
from autobib.util import find_in_path, get_bib_keys
import pytest

cwd = Path(__file__).parent
test_document_path = cwd / "data"


@pytest.mark.skipif(not find_in_path("latex"), reason="requires latex")
def test_autobib(tmpdir):
    for fn in ("main.tex", "bar.bib", "foo.bib"):
        shutil.copy(test_document_path / fn, tmpdir / fn)

    subp.run(["latex", "main.tex"], cwd=tmpdir)
    p = subp.run(["bibtex", "main"], cwd=tmpdir)
    assert p.returncode == 0

    with open(tmpdir / "main.bib") as f:
        assert get_bib_keys(f.read()) == {
            "2020Univ....6..102M",
            "Vanthieghem:2021akb",
            "PierreAuger:2021qsd",
        }
        # "Dembinski:2018ihc" is not in main.bib, because it is in foo.bib
        # "Baur:2019cpv" is not in main.bib, because it is in bar.bib

    assert subp.run(["bibtex", "foobarbaz"], cwd=tmpdir).returncode != 0


@pytest.mark.skipif(not find_in_path("latex"), reason="requires latex")
def test_autobib_updated_key(tmpdir):
    # test renamed key Abe:1993xy -> CDF:1993wpv

    with (tmpdir / "main.tex").open("w") as f:
        f.write(
            r"""
\documentclass{article}
\begin{document}
\cite{Abe:1993xy}
\bibliographystyle{plain}
\bibliography{main}
\end{document}
"""
        )

    with (tmpdir / "main.bib").open("w") as f:
        f.write("")

    subp.run(["latex", "main.tex"], cwd=tmpdir)
    assert (tmpdir / "main.aux").exists()
    p = subp.run(["bibtex", "main"], cwd=tmpdir, stderr=subp.PIPE)
    assert p.returncode == 1
    # autobib deletes the aux file in this case, to force latex to re-create it next time
    assert not (tmpdir / "main.aux").exists()

    with open(tmpdir / "main.bib") as f:
        assert get_bib_keys(f.read()) == {"CDF:1993wpv"}

    assert (
        b"""Error: Keys need update in LaTeX document
  Abe:1993xy -> CDF:1993wpv
"""
        in p.stderr
    )



from autobib import util
from pathlib import Path

cwd = Path(__file__).parent
test_document_path = cwd / "data"
aux_path = test_document_path / "main.aux"


def test_get_aux_bibdata():
    bib_files = util.get_aux_bibdata(aux_path)

    assert bib_files == [
        test_document_path / "main.bib",
        test_document_path / "foo.bib",
        test_document_path / "bar.bib",
    ]


def test_get_aux_keys():
    citations = util.get_aux_keys(aux_path)

    assert citations == {
        "2020Univ....6..102M",
        "Vanthieghem:2021akb",
        "Baur:2019cpv",
        "Aab:2021zfr",
        "Dembinski:2018ihc",
    }


def test_get_entry_online_1():
    key = "PierreAuger:2021qsd"

    entry = util.get_entry_online(key)
    assert key in entry

    assert util.get_entry_online("foobarbaz") == ""


def test_get_entry_online_2():
    key = "2011ApJ...737..103S"
    entry = util.get_entry_online(key)
    assert key in entry

    assert util.get_entry_online("123foobarbaz") == ""



from pathlib import Path
import re
from typing import Set, List
import os
import urllib.request
import urllib.parse
import json


def get_bib_keys(txt: str) -> Set[str]:
    return set(re.findall(r"@[a-zA-Z]+\{([^,]+),", txt))


def get_aux_bibdata(aux: Path) -> List[Path]:
    with open(aux) as f:
        # having more than one bibdata is an error
        m = re.search(r"\\bibdata{([^}]+)}", f.read())
    # split comma-separated entries
    names = m.group(1).split(",") if m else []
    # make unique while preserving order
    dir = aux.parent
    return [dir / f"{name}.bib" for name in dict.fromkeys(names)]


def get_aux_keys(aux: Path) -> Set[str]:
    with open(aux) as f:
        txt = f.read()
    tmp = re.findall(r"\\citation{([^}]+)}", txt)
    # split multi-citations generated by `cite` package in LaTeX
    result = set()
    for c in tmp:
        result.update(c.split(","))
    return result


def get_entry_online(key: str) -> str:
    # Inspire and ADS keys are supported. ADS keys start with digits.
    if re.match("[0-9]+", key):
        # https://github.com/adsabs/adsabs-dev-api
        token = os.environ.get("ADS_TOKEN", "")
        if not token:
            print(
                "Key looks like ADS format, but the ADS_TOKEN environment variable "
                "environment variable is not set. You need to follow these steps:\n"
                "\n"
                "1) Follow the instructions at "
                "https://github.com/adsabs/adsabs-dev-api#access to get an API token.\n"
                "2) Export the token in your shell as ADS_TOKEN with\n"
                "      export ADS_TOKEN=<insert token here>"
            )
        req = urllib.request.Request(
            "https://api.adsabs.harvard.edu/v1/export/bibtex",
            f'{{"bibcode": ["{key}"]}}'.encode("ascii"),
            {
                "Authorization": f"Bearer {token}",
                "Content-Type": "application/json",
            },
        )
        try:
            with urllib.request.urlopen(req) as r:
                # do not check return code, we accept failure silently
                bibdata = json.loads(r.read()).get("export", "\n")[:-1]
        except urllib.error.HTTPError:
            bibdata = ""
    else:
        # https://github.com/inspirehep/rest-api-doc
        data = urllib.parse.urlencode({"q": key, "format": "bibtex"})
        req = urllib.request.Request(f"https://inspirehep.net/api/literature?{data}")
        with urllib.request.urlopen(req) as r:
            # do not check return code, we accept failure silently
            bibdata = r.read().decode() if r.code == 200 else ""
    assert len(get_bib_keys(bibdata)) <= 1
    return bibdata


def find_in_path(name: str) -> List[Path]:
    paths = ["/bin", "/usr/bin", "/usr/local/bin"]
    for path in os.environ["PATH"].split(":"):
        path = os.path.expandvars(path)
        paths.append(path)

    results = []
    for path in paths:
        for p in Path(os.path.expandvars(path)).glob(name):
            results.append(p)

    return results



from ._version import version as __version__  # noqa


def main() -> int:
    import sys
    from pathlib import Path
    from . import util
    import subprocess as subp

    args = sys.argv[1:]

    def log(x: str) -> None:
        print(f"autobib: {x}")

    if "--version" in args:
        print(f"autobib {__version__}")
    if len(args) == 1:
        # bibtex accepts filename without extension
        aux = Path(args[0]).with_suffix(".aux")
        if aux.exists():
            log(f"Found {aux}")
            bib_files = util.get_aux_bibdata(aux)
            citations = util.get_aux_keys(aux)

            main = None
            all = set()
            for bib in bib_files:
                if bib.exists():
                    with open(bib) as f:
                        keys = util.get_bib_keys(f.read())
                        if main is None:
                            main = keys
                        all |= keys

            unknown = citations - all

            # Inspire is currently updating their keys. Entries are still found under
            # the old key, but an entry with the new key is returned.
            keys_need_update = []
            if unknown:
                bib = bib_files[0]
                with open(bib, "a") as f:
                    for c in unknown:
                        log(f"Fetching online: {c}")
                        ref = util.get_entry_online(c)
                        if not ref:
                            log(f"Warning: no entry found for '{c}'")
                            continue
                        start = ref.find("{") + 1
                        key = ref[start : ref.find(",", start)]
                        if key != c:
                            log(f"Warning: fetched key {key} differs from query {c}")
                            keys_need_update.append((c, key))
                            if key in all:  # already downloaded
                                continue
                        log(f"Writing {c} to {bib}")
                        f.write("\n")
                        f.write(ref)

            if keys_need_update:
                msg = ["Error: Keys need update in LaTeX document"]
                for c, key in keys_need_update:
                    msg.append(f"  {c} -> {key}")
                # delete the aux file, because it needs to be recreated in this case
                aux.unlink()
                raise SystemExit("\n".join(msg))

    # find original bibtex
    bibtexs = util.find_in_path("bibtex")
    assert len(bibtexs) > 1
    # check that we are first in path and skip all other instances of the bibtex script
    for i, bibtex in enumerate(bibtexs):
        with open(bibtex, "rb") as fb:
            is_script = fb.read(2) == b"#!"
        if is_script:
            continue
        else:
            assert i > 0
            break

    # run original bibtex
    return subp.run([bibtex] + args).returncode  # type: ignore
